/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */

/* Copyright (c) 2002-2005, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#include <stdio.h>
#include "mpid.h"
#include "vapi_header.h"
#include "viadev.h"
#include "viapriv.h"
#include "viutil.h"
#include "vbuf.h"
#include "viapacket.h"
#include "process/pmgr_client.h"
#include <unistd.h>
#include "vapi_const.h"
/* for supporting multiple HCAs/ports */
int num_ports = 2;
int num_hcas = 1;
int num_subchannels = 2;

int link_speed[MAX_SUBCHANNELS];

int striping_threshold;

int rdma_scheduling = RDMA_EVEN_STRIPING;
int sr_scheduling = SR_USE_FIRST;
int sr_rr_window = DEFAULT_RR_WINDOW;

/* low level VIA information (hidden from ADI2) */
viadev_info_t viadev;

char nullsbuffer, nullrbuffer;

/* The vbuf backlog queue */
viadev_backlog_queue_t *backlog_queue = NULL;

void 
viadev_unimplemented(char *s)
{
    vapidev_error_abort(GEN_EXIT_ERR, "Function %s not implemented", s);
}


/* to handle Send Q overflow, we maintain an extended send queue
 * above the HCA.  This permits use to have a virtually unlimited send Q depth
 * (limited by number of vbufs available for send)
 */


void viadev_ext_sendq_queue(viadev_connection_t * c, vbuf * v, int subchannel)
{
    v->desc.next = NULL;
    if (c->ext_sendq_head[subchannel] == NULL) {
	    c->ext_sendq_head[subchannel] = v;
    } else {
    	c->ext_sendq_tail[subchannel]->desc.next = v;
    }
    c->ext_sendq_tail[subchannel] = v;
}


/* dequeue and send as many as we can from the extended send queue
 * this is called in each function which may post send prior to it attempting
 * its send, hence ordering of sends is maintained
 */
void viadev_ext_sendq_send(viadev_connection_t * c, int subchannel)
{
    VAPI_hca_hndl_t hca = viadev.nic[subchannel/num_ports];
    vbuf *v;
    int result;

    while (c->send_wqes_avail[subchannel] && c->ext_sendq_head[subchannel]) {
    /* dequeue one vbuf */
    	v = c->ext_sendq_head[subchannel];
	    c->ext_sendq_head[subchannel] = v->desc.next;
    	if (v == c->ext_sendq_tail[subchannel]) {
        	c->ext_sendq_tail[subchannel] = NULL;
    	}
	    v->desc.next = NULL;
    	/* put the vbuf on the HCA's send Q */
	    c->send_wqes_avail[subchannel]--;
#ifdef USE_INLINE
    	if (v->desc.sg_entry.len > inline_size)
        	result = VAPI_post_sr(hca , c->vi[subchannel], &(v->desc.sr));
   	 	else
       	 	result =
        		EVAPI_post_inline_sr(hca, c->vi[subchannel], &(v->desc.sr));
#else
	    result = VAPI_post_sr(hca, c->vi[subchannel], &(v->desc.sr));
#endif
    	if (result != VAPI_OK) {
        	vapidev_error_abort(VAPI_RETURN_ERR,
           		     "VAPI_post_sr (viadev_ext_sendq_send)");
    	}
    }
}



/* init the backlog queue */
void viadev_init_backlog_queue(viadev_connection_t *c)
{
    int i;

    for (i = 0; i < num_subchannels; i++) {
	c->backlog[i].len = 0;
	c->backlog[i].vbuf_head = NULL;
	c->backlog[i].vbuf_tail = NULL;
    }
}

/* 
 * post a descriptor to the send queue
 * all outgoing packets go through this routine. 
 * takes a connection rather than a vi because we need to 
 * update flow control information on the connection. Also
 * it turns out it is always called in the context of a connection, 
 * i.e. it would be post_send(c->vi) otherwise. 
 * xxx should be inlined
 */

#define BACKLOG_ENQUEUE(q,v) {                     	\
    v->desc.next = NULL;                			\
    if (q->vbuf_tail == NULL) {                    	\
         q->vbuf_head = v;                         	\
    } else {                                       	\
         q->vbuf_tail->desc.next = v;   			\
    }                                              	\
    q->vbuf_tail = v;                              	\
    q->len++;                                      	\
}


/*add later */
#define BACKLOG_DEQUEUE(q,v)  {                    	\
    v = q->vbuf_head;                              	\
    q->vbuf_head = v->desc.next;        			\
    if (v == q->vbuf_tail) {                       	\
        q->vbuf_tail = NULL;                       	\
    }                                              	\
    q->len--;                                      	\
    v->desc.next = NULL;                			\
}

/* Just got a credit update, send off as many backlogged
 * messages as possible
 */

void viadev_backlog_send(viadev_connection_t *c, int subchannel)
{
    VAPI_hca_hndl_t hca = viadev.nic[subchannel/num_ports];
    viadev_backlog_queue_t *q = &(c->backlog[subchannel]);

    while ((q->len > 0) && (c->remote_credit[subchannel] > 0)) {
		vbuf *v = NULL;
		viadev_packet_header *p;
		int result;
		assert(q->vbuf_head != NULL);
		BACKLOG_DEQUEUE(q,v);

	/* Assumes packet header is at beginning of packet structure */
		p = (viadev_packet_header*)VBUF_BUFFER_START(v);
	
		PACKET_SET_CREDITS(p,c,subchannel);
		c->remote_credit[subchannel]--;

		v->grank = c->global_rank;
        if (!c->send_wqes_avail[subchannel]) {
            viadev_ext_sendq_queue(c, v, subchannel);
            continue;
        }
        c->send_wqes_avail[subchannel]--;

#ifdef USE_INLINE
		if (v->desc.sg_entry.len > inline_size)
	    	result = VAPI_post_sr(hca, c->vi[subchannel], &(v->desc.sr));
		else
	    	result = EVAPI_post_inline_sr(hca, c->vi[subchannel], &(v->desc.sr));
#else
   	    	result = VAPI_post_sr(hca, c->vi[subchannel], &(v->desc.sr));
#endif
    	if (result != VAPI_OK) {
        		vapidev_error_abort(VAPI_RETURN_ERR,
                	"VAPI_post_sr (viadev_backlog_push)");
    	}

    }
}

void viadev_post_send(viadev_connection_t *c, vbuf *v, int subchannel)
{
    int result;
    VAPI_hca_hndl_t hca = viadev.nic[subchannel/num_ports];

 	viadev_packet_header *p = 
			(viadev_packet_header*)	VBUF_BUFFER_START(v);
    
    if (c->remote_credit[subchannel] > 0 || p->type == VIADEV_PACKET_NOOP) {
	
	/* if we got here, the backlog queue better be  empty */
    	assert(c->backlog[subchannel].len == 0 || p->type == VIADEV_PACKET_NOOP);
    	D_PRINT("post_send: SENT [vi %d][%d][vbuf %p] %s",
    		    c->global_rank, c->remote_credit, v,
        		viadev_packet_to_string(c->vi, v));

		PACKET_SET_CREDITS(p,c,subchannel);

	    if (p->type != VIADEV_PACKET_NOOP)
        	c->remote_credit[subchannel]--;

		v->grank = c->global_rank;

		D_PRINT("send message: len=%d", v->desc.sg_entry.len);
		D_PRINT("vbuf=%p\n", v);
#if 1 
 
    if (!c->send_wqes_avail[subchannel]) {
        viadev_ext_sendq_queue(c, v, subchannel);
        return;
    }
    c->send_wqes_avail[subchannel]--;
#endif
#ifdef USE_INLINE
		if (v->desc.sg_entry.len > inline_size)
	    	result = VAPI_post_sr(hca, c->vi[subchannel], &(v->desc.sr));
		else
	    	result = EVAPI_post_inline_sr(hca, c->vi[subchannel], &(v->desc.sr));
#else
        	result = VAPI_post_sr(hca, c->vi[subchannel], &(v->desc.sr));
#endif	

    	if (result != VAPI_OK) {
        		vapidev_error_abort(VAPI_RETURN_ERR,
                	"VAPI_post_sr (viadev_post_send)");
    	}

    /* do post receive after post_sr for last part of message
     * so latency hidden during post_sr
     */

    	if (viadev_prepost_threshold && v->shandle != NULL
        	&& c->initialized[subchannel]) {
        	int needed =
        	viadev_prepost_depth + viadev_prepost_noop_extra +
        	MIN(viadev_prepost_rendezvous_extra,
            	c->rendezvous_packets_expected);
        	while (c->preposts[subchannel] < viadev_rq_size && c->preposts[subchannel] < needed) {
        		PREPOST_VBUF_RECV(c, subchannel);
        	}
        	viadev_send_noop_ifneeded(c, subchannel);
    	}
    } else {
	/* must delay the send until more credits arrive.
	 * put on end of queue
	 */
	viadev_backlog_queue_t *q = &c->backlog[subchannel];
	BACKLOG_ENQUEUE(q,v);

    D_PRINT("post_send QUEUED: [vi %d][%d %d %d][vbuf %p] %s",
        c->global_rank, c->remote_credit, c->local_credit,
        c->remote_cc, v, viadev_packet_to_string(c->vi, v));

    }

}


#if defined(RDMA_FAST_PATH)
#ifdef __ia64__
#define FAST_RDMA_CAL_PADDING(_len)                          {        	\
      VBUF_FLAG_TYPE tem;                                               \
      tem = _len + (0x8 - ((_len + VBUF_FAST_RDMA_EXTRA_BYTES) & 0x7)); \
      if( tem < VBUF_BUFFER_SIZE)                                       \
              _len = tem;                                               \
}
#else
#define FAST_RDMA_CAL_PADDING(len)
#endif
#endif

#if defined(RDMA_FAST_PATH)

void post_fast_rdma(viadev_connection_t *c, int len, 
					int subchannel)
{
    int n = c->phead_RDMA_send[subchannel];
    VBUF_FLAG_TYPE flag;

    int result;
    viadev_packet_header *p = (viadev_packet_header*)
		(c->RDMA_send_buf[subchannel][n].buffer);

    VAPI_hca_hndl_t hca = viadev.nic[subchannel/num_ports];

    /* set credits info */
    PACKET_SET_CREDITS(p, c, subchannel);
   
    /* advance the head pointer */
    if (++(c->phead_RDMA_send[subchannel]) >= viadev_num_rdma_buffer)
	    c->phead_RDMA_send[subchannel] = 0;

    FAST_RDMA_CAL_PADDING(len);

    /* set length, 3 extra ints for flags and size */
    c->RDMA_send_buf[subchannel][n].desc.sg_entry.len =
  	  len + VBUF_FAST_RDMA_EXTRA_BYTES;
    /* set flags */
    if (*(VBUF_FLAG_TYPE *) (c->RDMA_send_buf[subchannel][n].buffer + len) == len) {
    	flag = (VBUF_FLAG_TYPE) (len + FAST_RDMA_ALT_TAG);
    } else {
	    flag = (VBUF_FLAG_TYPE) len;
    }

    /* set head flag */
    c->RDMA_send_buf[subchannel][n].head_flag = flag;

    /* set tail flag */
    *(VBUF_FLAG_TYPE *)(c->RDMA_send_buf[subchannel][n].buffer + len) = flag;

    c->num_no_completion[subchannel]++;
    if (c->num_no_completion[subchannel] >= MAX_NO_COMPLETION) {
	    c->num_no_completion[subchannel] = 0;
	    /* generate a completion */

	c->RDMA_send_buf[subchannel][n].desc.sr.comp_type = VAPI_SIGNALED;
	c->RDMA_send_buf[subchannel][n].desc.sr.id = (aint_t) 
		&(c->RDMA_send_buf[subchannel][n]); 
    } 

    if (!c->send_wqes_avail[subchannel]) {
    	viadev_ext_sendq_queue(c, &c->RDMA_send_buf[subchannel][n], subchannel);
	    return;
    }
    c->send_wqes_avail[subchannel]--;

#ifdef USE_INLINE
    if (len > inline_size)
	result = VAPI_post_sr(hca, c->vi[subchannel], 
			&(c->RDMA_send_buf[subchannel][n].desc.sr));
    else
	result = EVAPI_post_inline_sr(hca, c->vi[subchannel], 
			&(c->RDMA_send_buf[subchannel][n].desc.sr));
#else
    result = VAPI_post_sr(hca, c->vi[subchannel], 
			&(c->RDMA_send_buf[subchannel][n].desc.sr));
#endif

    if (result != VAPI_OK) {
    vapidev_error_abort(VAPI_RETURN_ERR,
                "VAPI_post_sr (post_fast_rdma)");
    }


}


void post_fast_rdma_with_completion(viadev_connection_t *c, int len, int subchannel)
{

    int n = c->phead_RDMA_send[subchannel];
    VBUF_FLAG_TYPE flag;
    int result;
    viadev_packet_header *p = (viadev_packet_header*)
		(c->RDMA_send_buf[subchannel][n].buffer);
    VAPI_hca_hndl_t hca = viadev.nic[subchannel/num_ports];

    /* Mark RDMA buffer busy */
    c->RDMA_send_buf[subchannel][n].padding = BUSY_FLAG;

#ifndef DISABLE_HEADER_CACHING
#define MAX_FAST_EAGER_SIZE 255
    if ((p->type != FAST_EAGER_CACHED) &&
    	(p->type != VIADEV_PACKET_EAGER_START))
	    PACKET_SET_CREDITS(p, c, subchannel);
    if ((p->type == VIADEV_PACKET_EAGER_START) &&
    	(len >= MAX_FAST_EAGER_SIZE))
    	PACKET_SET_CREDITS(p, c, subchannel);
#else
	    PACKET_SET_CREDITS(p, c, subchannel);
#endif
    if (++(c->phead_RDMA_send[subchannel]) >= viadev_num_rdma_buffer)
	    c->phead_RDMA_send[subchannel] = 0;
    FAST_RDMA_CAL_PADDING(len);

    c->RDMA_send_buf[subchannel][n].desc.sg_entry.len =
    len + VBUF_FAST_RDMA_EXTRA_BYTES;

    /* set flags */
    if (*(VBUF_FLAG_TYPE *) (c->RDMA_send_buf[subchannel][n].buffer + len)
	    == len) {
    	flag = (VBUF_FLAG_TYPE) (len + FAST_RDMA_ALT_TAG);
    } else {
	    flag = (VBUF_FLAG_TYPE) len;
    }
    /* set head flag */
    c->RDMA_send_buf[subchannel][n].head_flag = flag;

    /* set tail flag */
    *(VBUF_FLAG_TYPE *)(c->RDMA_send_buf[subchannel][n].buffer + len) = flag;

    c->num_no_completion[subchannel] = 0;
    /* generate a completion */
    c->RDMA_send_buf[subchannel][n].desc.sr.comp_type = VAPI_SIGNALED;
    c->RDMA_send_buf[subchannel][n].desc.sr.id = (aint_t) 
			&(c->RDMA_send_buf[subchannel][n]); 

    if (!c->send_wqes_avail[subchannel]) {
        viadev_ext_sendq_queue(c, &c->RDMA_send_buf[subchannel][n], subchannel);
        return;
    }
    c->send_wqes_avail[subchannel]--;

#ifdef _IA64_
    len += 16;
    len -= (len & 7);
    if (len > VBUF_BUFFER_SIZE)
        len = VBUF_BUFFER_SIZE;
#endif

    c->RDMA_send_buf[subchannel][n].desc.sg_entry.len =
    len + VBUF_FAST_RDMA_EXTRA_BYTES;

#ifdef USE_INLINE
    if (len + VBUF_FAST_RDMA_EXTRA_BYTES> inline_size)
    result = VAPI_post_sr(hca, c->vi[subchannel], 
						&(c->RDMA_send_buf[subchannel][n].desc.sr));
    else
    result = EVAPI_post_inline_sr(hca, c->vi[subchannel], 
						&(c->RDMA_send_buf[subchannel][n].desc.sr));
#else
    result = VAPI_post_sr(hca, c->vi[subchannel], 
						&(c->RDMA_send_buf[subchannel][n].desc.sr));
#endif


}



inline int fast_rdma_ok(viadev_connection_t *c, 
						int data_size, int subchannel)
{

    /* not OK if buffers are full */

    if (viadev_num_rdma_buffer == 0)
    return 0;

    if (c->phead_RDMA_send[subchannel] == 
		c->ptail_RDMA_send[subchannel]) {
    return 0;
    }

    /* not OK if data size is too large */
    if (data_size > VBUF_BUFFER_SIZE) {	
		return 0;
	}
    /* not OK if there are sends waiting in the backlog ??? */
    if (c->backlog[subchannel].len > 0){
		 return 0;
	}
    /* See if head buffer is marked free */
    if(c->RDMA_send_buf[subchannel][c->phead_RDMA_send[subchannel]].padding == BUSY_FLAG) {
        return 0;
    }
    /* otherwise OK */
    return 1;
}

inline void release_recv_rdma(viadev_connection_t *c, vbuf *v)
{
    vbuf *next_free;
    int next;
    int i;
    int subchannel = v->subchannel;

#if 1 
    next = (c->p_RDMA_recv_tail[subchannel] + 1) % viadev_num_rdma_buffer;
    next_free = &(c->RDMA_recv_buf[subchannel][next]);
#else
    next = c->p_RDMA_recv_tail[subchannel] + 1;
    if (next >= viadev_num_rdma_buffer)
    	next = 0;
    next_free = &(c->RDMA_recv_buf[subchannel][next]);

#endif
    
    v->padding = FREE_FLAG;
    if (v != next_free) {
		return;
    }
#if 1 
    /* search all free buffers */
    for (i = next; i != c->p_RDMA_recv[subchannel]; i = (i + 1) % viadev_num_rdma_buffer) {
	if (c->RDMA_recv_buf[subchannel][i].padding == FREE_FLAG) {
	    c->rdma_credit[subchannel]++;
	    c->p_RDMA_recv_tail[subchannel] = (c->p_RDMA_recv_tail[subchannel] + 1) % viadev_num_rdma_buffer;

	    c->RDMA_recv_buf[subchannel][i].padding = BUSY_FLAG;
    } else
		 break;
	}
#else
    /* search all free buffers */
    for (i = next; i != c->p_RDMA_recv[subchannel];) {

    if (c->RDMA_recv_buf[subchannel][i].padding == FREE_FLAG) {
        c->rdma_credit[subchannel]++;
        if (++(c->p_RDMA_recv_tail[subchannel]) >= viadev_num_rdma_buffer)
        c->p_RDMA_recv_tail[subchannel] = 0;
        c->RDMA_recv_buf[i][subchannel].padding = BUSY_FLAG;
    } else
		 break;
    if (++i >= viadev_num_rdma_buffer)
        i = 0;
    }
#endif
}


#ifndef DISABLE_HEADER_CACHING

int search_header_cache(viadev_connection_t * c,
	                    viadev_packet_eager_start * h)
{
    viadev_packet_eager_start *cached = &(c->cached_outgoing);

    if ((cached->header.vbuf_credit == h->header.vbuf_credit) &&
	    (cached->header.remote_credit == h->header.remote_credit) &&
    	(cached->header.rdma_credit == h->header.rdma_credit) &&
	    (cached->envelope.context == h->envelope.context) &&
    	(cached->envelope.tag == h->envelope.tag) &&
	    (cached->envelope.src_lrank == h->envelope.src_lrank)) {
    	c->cached_hit++;
    return 1;
    }

    c->cached_miss++;
    memcpy(cached, h, sizeof(viadev_packet_eager_start));
    return 0;
}

#endif

#endif

void viadev_post_rdmawrite(	viadev_connection_t *c, 
							vbuf *v, int subchannel)
{
    int result;
    VAPI_hca_hndl_t hca = viadev.nic[subchannel/num_ports];


    D_PRINT("post_rdmawrite: [vi %d] %s",
		     c->global_rank, viadev_packet_to_string(c->vi, v));

    /* set id to be the global_rank */
    v->grank = c->global_rank;

#if 1
    if (!c->send_wqes_avail[subchannel]) {
        viadev_ext_sendq_queue(c, v, subchannel);
        return;
    }
    c->send_wqes_avail[subchannel]--;
#endif

#ifdef USE_INLINE
    if (v->desc.sg_entry.len > inline_size)
	result = VAPI_post_sr(hca, c->vi[subchannel], &(v->desc.sr));
    else
	result = EVAPI_post_inline_sr(hca, c->vi[subchannel], &(v->desc.sr));
#else
    result = VAPI_post_sr(hca, c->vi[subchannel], &(v->desc.sr));
#endif

    if(result != VAPI_OK) {
        vapidev_error_abort(VAPI_RETURN_ERR, "VAPI_post_sr");
    }

}

/* 
 * post a descriptor to the receive queue
 */

void viadev_post_recv(viadev_connection_t *c, vbuf *v, int subchannel)
{
    int result;
    VAPI_hca_hndl_t hca = viadev.nic[subchannel/num_ports];

    D_PRINT("viadev_post_recv: v = " AINT_FORMAT, (aint_t) v);

    /* set id to be the global_rank */
    v->grank = c->global_rank;

	if(v->grank != viadev.me) {
	    result = VAPI_post_rr(hca, c->vi[subchannel], &(v->desc.rr));

    	if (result != VAPI_OK) {
    		vapidev_error_abort(VAPI_RETURN_ERR,
                "VAPI_post_rr (viadev_post_recv)");
    	}
	}
}

char *viadev_packet_to_string(VAPI_qp_hndl_t vi, vbuf *v)
{
	return '\0';
}


int vi_to_grank(VAPI_qp_hndl_t vi)
{
    return 0;
}

#ifdef MCST_SUPPORT
void viadev_post_ud_recv(vbuf *);

void viadev_post_ud_recv(vbuf * v)
{

    int result;
    v->grank = -1;

    D_PRINT("viadev_post_ud_recv: v = " AINT_FORMAT, (aint_t) v);
    result = VAPI_post_rr(viadev.nic[NIC_0], viadev.ud_qp_hndl, &(v->desc.rr));
    if (result != VAPI_OK) {
        vapidev_error_abort(VAPI_RETURN_ERR,
                            "VAPI_post_rr (viadev_post_recv)");
    }
}

#endif


int choose_sr_subchannel(viadev_connection_t *c)
{
    return 0;
}

/* striping rdma messages across multiple subchannels */
void schedule_rdma_even_striping(viadev_connection_t *c, char *addr, int size,
	char** stripe_start_addr, int *stripe_size)
{
    /* do even striping */
    int i;
    char *next_addr;
    int inc;
    if (size < striping_threshold) {
	/* use the first connection for small messages */
	stripe_start_addr[0] = addr;
	stripe_size[0] = size;
	for (i = 1; i < num_subchannels; i++) stripe_size[i] = 0;
#if 0
	stripe_start_addr[1] = addr;
	stripe_size[1] = size;
	stripe_size[0] = 0;
#endif
    } else {
	/* strip evenly across all connections */
	D_PRINT("stripping triggered\n");
	inc = size / num_subchannels;
	next_addr = addr;
	for (i = 0; i < num_subchannels - 1; i++) {
	    stripe_start_addr[i] = next_addr;
	    stripe_size[i] = inc;
	    next_addr += inc;
	}
	stripe_start_addr[num_subchannels - 1] = next_addr;
	stripe_size[num_subchannels - 1] = size - (num_subchannels - 1) * inc;
    }

}

void schedule_rdma(viadev_connection_t *c, char *addr, int size,
	char** stripe_start_addr, int *stripe_size)
{
    schedule_rdma_even_striping(c, addr, size, stripe_start_addr, stripe_size);
}


/* check ofo queues of all connections to see if
   there is any packet (vbuf) that matches the current
   expected sequence number 
   */
int check_ofo_queues(vbuf **vbuf_addr)
{
    int i;
    viadev_packet_header *h;
    vbuf *v;
    
    for (i = 0; i < viadev.np; i++) {
	viadev_connection_t *c = &viadev.connections[i];
	if (i == viadev.me) continue;

	/* check the ofo queue */
	v = c->ofo_head;
	while (v) {
	    h = (viadev_packet_header *)VBUF_BUFFER_START(v);

	    /* in-order packet? */
	    if (h->id == c->next_packet_expected) {
		remove_ofo_queue(v, c);
		*vbuf_addr = v;
		return VAPI_OK;
	    }

	    v = (vbuf *)(v->desc.next);
	}
	
    }

    return VAPI_CQ_EMPTY;

}

/* add a packet (vbuf) to an ofo queue */
void enqueue_ofo_queue(vbuf *vbuf_addr, viadev_connection_t *c)
{
    vbuf_addr->desc.next = c->ofo_head;
    c->ofo_head = vbuf_addr;
}

/* remove a vbuf from an ofo queue. The vbuf has
   to be in the queue already 
   */
void remove_ofo_queue(vbuf *vbuf_addr, viadev_connection_t *c)
{
    vbuf *v, *v_next;
 
    v = c->ofo_head;
    if (v == NULL) {
	fprintf(stderr, "remove_ofo_queue error, queue empty! vbuf = %p\n", vbuf_addr);
	exit(-1);
    }

    if (v == vbuf_addr) {
	c->ofo_head = (vbuf *)(v->desc.next);
	return;
    }

    v_next = (vbuf *)(v->desc.next);
    while (v_next) {
	if (v_next == vbuf_addr) {
	    v->desc.next = v_next->desc.next;
	    return;
	}
	v = v_next;
	v_next = (vbuf *)(v_next->desc.next);
    }
    
    fprintf(stderr, "remove_ofo_queue error, not found! vbuf = %p\n", vbuf_addr);
    exit(-1);

}

/* init all ofo queues to NULL */
void init_ofo_queues(void)
{
    int i;
    for (i = 0; i < viadev.np; i++) {
	viadev_connection_t *c = &viadev.connections[i];
	c->ofo_head = NULL;
    }
}
